import sys
import yaml

from intake.cli.bootstrap import main as run
from intake.cli.util import Subcommand

from .catalog import DCATCatalog
from .util import mirror_data


class Mirror(Subcommand):
    """
    Mirror a catalog subset specified in a manifest to a remote bucket,
    and print the resulting subsetted catalog to stdout.


    If --dry-run is specified, then no upload happens, but the catalog is still printed.
    """

    name = "mirror"

    def initialize(self):
        """
        Initialize the subcommand by adding the argparser arguments.
        """
        self.parser.add_argument(
            "manifest",
            type=str,
            metavar="MANIFEST",
            help="Path to a manifest YAML file",
        )
        self.parser.add_argument(
            "--dry-run",
            help="If this flag is given, no upload occurs",
            action="store_true",
        )
        self.parser.add_argument(
            "--version", metavar="VERSION", type=str, help="Catalog version"
        )
        self.parser.add_argument(
            "--name", metavar="VERSION", type=str, help="Catalog name"
        )

    def invoke(self, args):
        """
        Invoke the command.
        """
        upload = not args.dry_run
        catalog = mirror_data(
            args.manifest, upload=upload, name=args.name, version=args.version
        )
        print(yaml.dump(catalog))


class Create(Subcommand):
    """
    Create an intake YAML catalog from a DCAT catalog and print it to stdout.
    """

    name = "create"

    def initialize(self):
        """
        Initialize the subcommand by adding the argparser arguments.
        """
        self.parser.add_argument("uri", metavar="URI", type=str, help="Catalog URI")
        self.parser.add_argument(
            "--version", metavar="VERSION", type=str, help="Catalog version"
        )
        self.parser.add_argument(
            "--name", metavar="VERSION", type=str, help="Catalog name"
        )

    def invoke(self, args):
        """
        Invoke the command.
        """
        version = args.version or None
        name = args.name or ""
        metadata = {"name": name, "version": version}
        catalog = DCATCatalog(args.uri, name=name, metadata=metadata)
        print(catalog.serialize())


subcommands = [Mirror, Create]


def main(argv=None):
    return run("Intake DCAT CLI", subcommands, argv or sys.argv)


if __name__ == "__main__":
    sys.exit(main(sys.argv))



from typing import ClassVar

import requests
import yaml

from intake.catalog import Catalog
from intake.catalog.local import LocalCatalogEntry

from .distributions import get_relevant_distribution, DEFAULT_PRIORITIES
from . import _version


class DCATCatalog(Catalog):
    """
    A Catalog that references a DCAT catalog at some URL
    and constructs an intake catalog from it, with opinionated
    choices about the drivers that will be used to load the datasets.
    In general, the drivers are in order of decreasing specificity:

    GeoJSON
    Shapefile
    CSV
    """

    name: ClassVar[str] = "dcat"
    version: ClassVar[str] = _version

    def __init__(
        self, url, name="catalog", items=None, priority=None, metadata=None, **kwargs
    ):
        """
        Initialize the catalog.

        Parameters
        ----------
        url: str
            A URL pointing to a DCAT catalog, usually named data.json
        name: str
            A name for the catalog
        items: Dict[str, str]
            A mapping of {name: id} of entries to include.
            The name is a human-readable name, and the id is the DCAT identifier.
            These items will be available in the catalog under the human-readable name.
            If `None` is given, then all entries are included, and keyed using
            their DCAT identifier.
        priority: List[str]
            A priority list for which drivers to choose when loading datasets.
            If none is give, it defaults to GeoJSON, followed by Shapefile,
            followed by CSV. This can also be used to remove options from the
            default list.
        metadata: dict
            Additional information about the catalog
        """
        self.url = url
        self.name = name
        self._items = items
        self._priority = priority
        super().__init__(name=name, metadata=metadata, **kwargs)

    def _load(self):
        """
        Load the catalog from the remote data source.
        """

        if self._priority:
            if not all(p in DEFAULT_PRIORITIES for p in self._priority):
                raise ValueError("Invalid list of priorities")

        resp = requests.get(self.url)
        catalog = resp.json()
        if self._items is not None:
            self._entries = {}
            for key, identifier in self._items.items():
                entry = next(
                    (e for e in catalog["dataset"] if e["identifier"] == identifier),
                    None,
                )
                if entry and _should_include_entry(entry, self._priority):
                    self._entries[key] = DCATEntry(entry, self._priority)
        else:
            self._entries = {
                entry["identifier"]: DCATEntry(entry, self._priority)
                for entry in catalog["dataset"]
                if _should_include_entry(entry, self._priority)
            }

    def serialize(self):
        """
        Serialize the catalog to yaml.

        Returns
        -------
        A string with the yaml-formatted catalog.
        """
        output = {"metadata": self.metadata, "sources": {}}
        for key, entry in self.items():
            output["sources"][key] = yaml.safe_load(entry.yaml())["sources"][key]
        return yaml.dump(output)


class DCATEntry(LocalCatalogEntry):
    """
    A class representign a DCAT catalog entry, which knows how to pretty-print itself.
    """

    def __init__(self, dcat_entry, priority=None):
        """
        Construct an Intake catalog entry from a DCAT catalog entry.
        """
        driver, args = get_relevant_distribution(dcat_entry, priority)
        name = dcat_entry["identifier"]
        description = f"## {dcat_entry['title']}\n\n{dcat_entry['description']}"
        metadata = {"dcat": dcat_entry}
        super().__init__(name, description, driver, True, args=args, metadata=metadata)

    def _ipython_display_(self):
        """
        Print an HTML repr for the entry
        """
        dcat = self.metadata["dcat"]
        title = dcat.get("title") or "unknown"
        entry_id = dcat.get("identifier")
        description = dcat.get("description")
        issued = dcat.get("issued") or "unknown"
        modified = dcat.get("modified") or "unknown"
        license = dcat.get("license") or "unknown"
        organization = dcat.get("publisher")
        publisher = organization.get("name") or "unknown" if organization else "unknown"
        download = self._open_args.get("urlpath") or "unknown"

        info = f"""
            <p style="margin-bottom: 0.5em"><b>ID:</b><a href="{entry_id}"> {entry_id}</a></p>
            <p style="margin-bottom: 0.5em"><b>Issued:</b> {issued}</p>
            <p style="margin-bottom: 0.5em"><b>Last Modified:</b> {modified}</p>
            <p style="margin-bottom: 0.5em"><b>Publisher:</b> {publisher}</p>
            <p style="margin-bottom: 0.5em"><b>License:</b> {license}</p>
            <p style="margin-bottom: 0.5em"><b>Download URL:</b><a href="{download}"> {download}</a></p>
        """  # noqa: E501
        html = f"""
        <h3>{title}</h3>
        <div style="display: flex; flex-direction: row; flex-wrap: wrap; height:256px">
            <div style="flex: 0 0 384px; padding-right: 24px">
                {info}
            </div>
            <div style="flex: 1 1 0; height: 100%; overflow: auto">
                <p>
                    {description}
                </p>
            </div>
        </div>
        """

        return display(  # noqa: F821
            {
                "text/html": html,
                "text/plain": "\n".join([entry_id, title, description]),
            },
            raw=True,
        )


def _should_include_entry(dcat_entry, priority=None):
    """
    Return if a given DCAT entry should be included in the dataset.
    Returns True if we can find a driver to load it (GeoJSON,
    Shapefile, CSV), False otherwise.
    """
    return get_relevant_distribution(dcat_entry, priority) is not None



import copy
import time

import requests
import yaml

from dask.utils import tmpfile
import s3fs

from .catalog import DCATCatalog

# TODO: should we allow the user to pass in a s3fs session here?
fs = s3fs.S3FileSystem()

MAX_DOWNLOAD_ATTEMPTS = 60
RETRY_DELAY = 5


def mirror_data(manifest_file, upload=True, name=None, version=None):
    """
    Given a path the a manifest.yml file, download the relevant data,
    upload it to the specified bucket, and return a new catalog
    pointing at the data.

    Parameters
    ----------
    manifest_file: str
        A path to a manifest file.

    upload: boolean
        Whether to upload the datasets to the indicated bucket. Defaults to
        True, but can be set to false to perform a dry run.

    Returns
    -------
    A dictionary containing data for the new catalog.
    """
    new_catalog = {"metadata": {"name": name, "version": version}, "sources": {}}
    with open(manifest_file) as f:
        manifest = yaml.safe_load(f)
        bucket_uri = manifest["metadata"]["bucket_uri"]
        for catalog_name, catalog_data in manifest["sources"].items():
            catalog = DCATCatalog(catalog_data["args"]["url"], name=catalog_name)
            items = catalog_data["args"]["items"]
            for name, id in items.items():
                entry = yaml.safe_load(catalog[id].yaml())["sources"][id]
                new_entry = _construct_remote_entry(
                    bucket_uri, entry, name, upload=upload
                )
                new_catalog["sources"][name] = new_entry

    return new_catalog


def _upload_remote_data(old_uri, new_uri, dir=None):
    # Some servers, such as ESRI open data, can return 202 Accepted status codes
    # while it is preparing a dataset. In that case, we enter a retry loop.
    attempts = 0
    while attempts < MAX_DOWNLOAD_ATTEMPTS:
        r = requests.get(old_uri)
        if r.status_code == 202:
            attempts += 1
            time.sleep(RETRY_DELAY)
            continue
        elif r.status_code == 200:
            break
        else:
            raise RuntimeError(f"Unexpected response from DCAT server: {r.content}")
    else:
        raise RuntimeError("Unable to download dataset: Maximum retries met.")

    with tmpfile(dir=dir) as filename:
        with open(filename, "wb") as outfile:
            outfile.write(r.content)
        fs.put(filename, new_uri)


def _construct_remote_entry(bucket_uri, entry, name, directory="", upload=True):
    new_entry = copy.deepcopy(entry)
    old_uri = entry["args"]["urlpath"]
    upload_uri = _construct_remote_uri(bucket_uri, entry, name, directory)
    compression = _get_compression_for_entry(entry)
    access_uri = f"{compression}+{upload_uri}" if compression else upload_uri
    new_entry["args"]["urlpath"] = access_uri
    if upload:
        _upload_remote_data(old_uri, upload_uri)
    return new_entry


def _construct_remote_uri(bucket_uri, entry, name, directory=""):
    ext = _get_extension_for_entry(entry)
    key = f"{directory.strip('/')}/{name}{ext}" if directory else f"{name}{ext}"
    return f"{bucket_uri.strip('/')}/{key}"


def _get_extension_for_entry(intake_entry):
    """
    Given an intake catalog entry, return a file extension for it, which can
    be used to construct names when re-uploading the files to s3. It would be
    nice to be able to rely on extensions in the URL, but that is not
    particularly reliable. Instead, we infer an extension from the driver and
    args that are used. The following extensions are returned:

        GeoJSON: ".geojson"
        Zipped Shapefile: ".zip"
        Shapefile: ".shp"
        CSV: ".csv"
    """
    driver = intake_entry.get("driver")
    if driver == "geojson" or driver == "intake_geopandas.geopandas.GeoJSONSource":
        return ".geojson"
    elif (
        driver == "shapefile" or driver == "intake_geopandas.geopandas.ShapefileSource"
    ):
        return ".zip"
    elif driver == "csv" or driver == "intake.source.csv.CSVSource":
        return ".csv"
    else:
        raise ValueError(f"Unsupported driver {driver}")


def _get_compression_for_entry(intake_entry):
    """
    Given an intake catalog entry, determine a compression scheme, if
    applicable. Fiona uses compound protocols like 'zip+s3://' to construct
    paths for GDAL, and we need to construct that properly.

    Returns a string scheme like "zip", "gzip",  or "tar". If no compression
    scheme is used, returns `None`.
    """
    driver = intake_entry.get("driver")
    if driver == "shapefile" or driver == "intake_geopandas.geopandas.ShapefileSource":
        return "zip"
    else:
        return None



"""The setup script."""
from setuptools import setup, find_packages
from os.path import exists

from intake_dcat._version import __version__ as version

readme = open("README.md").read() if exists("README.md") else ""


setup(
    name="intake-dcat",
    description="DCAT to Intake Catalog translation layer",
    long_description=readme,
    long_description_content_type="text/markdown",
    maintainer="Ian Rose",
    maintainer_email="ian.rose@lacity.org",
    url="https://github.com/CityOfLosAngeles/intake-dcat",
    packages=find_packages(),
    package_dir={"intake-dcat": "intake-dcat"},
    include_package_data=True,
    install_requires=[
        "geopandas>=0.5",
        "intake>=0.5",
        "intake-geopandas>=0.2.2",
        "pyyaml",
        "requests",
        "s3fs",
    ],
    entry_points={
        "console_scripts": ["intake-dcat = intake_dcat.cli:main"],
        "intake.drivers": ["dcat = intake_dcat.catalog:DCATCatalog"],
    },
    license="Apache-2.0 license",
    zip_safe=False,
    keywords="intake dcat",
    version=version,
)



import re

DEFAULT_PRIORITIES = ["geojson", "shapefile", "csv"]


def get_relevant_distribution(dcat_entry, priority=None):
    """
    Given a DCAT entry, find the most relevant distribution
    for the intake catalog. Returns a tuple of
    (intake_driver_name, distribution). In general,
    we choose the more specific formats over the less specific
    formats. By default, they are ranked in the following order:

        GeoJSON
        Shapefile
        CSV

    You can provide a "priority" list to customize the priority.

    If none of these are found, None.
    If there are no distributions, it returns None.
    """
    distributions = dcat_entry.get("distribution")
    if not distributions or not len(distributions):
        return None

    priority = priority or DEFAULT_PRIORITIES
    for p in priority:
        if p.lower() == "geojson":
            for d in distributions:
                if test_geojson(d):
                    return "geojson", geojson_driver_args(d)
        elif p.lower() == "shapefile":
            for d in distributions:
                if test_shapefile(d):
                    return "shapefile", shapefile_driver_args(d)
        elif p.lower() == "csv":
            for d in distributions:
                if test_csv(d):
                    return "csv", csv_driver_args(d)
        else:
            raise ValueError(f"Unexpected driver {p}")

    return None


def test_geojson(distribution):
    """
    Test if a DCAT:distribution is GeoJSON.
    """
    return (
        distribution.get("mediaType") == "application/vnd.geo+json"
        or distribution.get("format", "").lower() == "geojson"
    )


def test_csv(distribution):
    """
    Test if a DCAT:distribution is CSV.
    """
    return (
        distribution.get("mediaType") == "text/csv"
        or distribution.get("format", "").lower() == "csv"
    )


def test_shapefile(distribution):
    """
    Test if a DCAT:distribution is a Shapefile.
    """
    # TODO: can there be a more robust test here?
    url = distribution.get("downloadURL") or distribution.get("accessURL") or ""
    title = distribution.get("title", "").lower()
    format = distribution.get("format", "").lower()
    return (
        title == "shapefile"
        or format == "shapefile"
        or re.search("format=shapefile", url, re.I) is not None
    )


def geojson_driver_args(distribution):
    """
    Construct driver args for a GeoJSON distribution.
    """
    url = distribution.get("downloadURL") or distribution.get("accessURL")
    if not url:
        raise KeyError(f"A download URL was not found for {str(distribution)}")
    return {"urlpath": url}


def csv_driver_args(distribution):
    """
    Construct driver args for a GeoJSON distribution.
    """
    url = distribution.get("downloadURL") or distribution.get("accessURL")
    if not url:
        raise KeyError(f"A download URL was not found for {str(distribution)}")
    return {"urlpath": url, "csv_kwargs": {"blocksize": None, "sample": False}}


def shapefile_driver_args(distribution):
    """
    Construct driver args for a GeoJSON distribution.
    """
    url = distribution.get("downloadURL") or distribution.get("accessURL")
    if not url:
        raise KeyError(f"A download URL was not found for {str(distribution)}")
    args = {"urlpath": url}
    return args
